import typing
from pathlib import Path
from typing import Union

from solara.kitchensink import react, sol, w

if typing.TYPE_CHECKING:
    import numpy as np


@react.component
def Image(image: Union[str, Path, "np.ndarray"]):
    if isinstance(image, (str, Path)):
        path = Path(image)
        if path.exists():
            import ipywidgets

            format = ipywidgets.Image._guess_format("image", str(path))
            value = ipywidgets.Image._load_file_value(str(path))
            return w.Image(value=value, format=format)
        elif isinstance(image, str):
            url_data = image.encode("utf8")
            return w.Image(value=url_data, format="url")
    elif sol.util.isinstanceof(image, "numpy:ndarray"):
        value = sol.util.numpy_to_image(image, format="png")
        return w.Image(value=value, format="image/png")
    raise TypeError(f"Only support URL, path or numpy array, not {image}")



import os
import traceback
from collections import abc

from rich import print


def start_error(title, msg, exception: Exception = None):
    if exception:
        traceback.print_exception(None, exception, exception.__traceback__)
    print(f"[red]{title}:\n\t[blue]{msg}")
    os._exit(-1)


def nested_get(object, dotted_name: str):
    names = dotted_name.split(".")
    for name in names:
        if isinstance(object, abc.Mapping):
            object = object.get(name)
        else:
            object = getattr(object, name)
    return object



from typing import Callable, List, TypeVar

from solara.kitchensink import react, v

T = TypeVar("T")


def _get_button_value(button: react.core.Element):
    value = button.kwargs.get("value")
    if value is None:
        value = button.kwargs.get("label")
    if value is None and button.args:
        value = button.args[0]
    return value


@react.component
def ToggleButtonsSingle(value: T, children: List[react.core.Element] = [], on_value: Callable[[T], None] = None):
    values = [_get_button_value(button) for button in children]
    index, set_index = react.use_state(values.index(value), key="index")

    def on_index(index):
        set_index(index)
        value = values[index]
        if on_value:
            on_value(value)

    with v.BtnToggle(children=children, multiple=False, mandatory=True, v_model=index, on_v_model=on_index) as main:
        pass
    return main


@react.component
def ToggleButtonsMultiple(value: List[T], children: List[react.core.Element] = [], on_value: Callable[[List[T]], None] = None):
    allvalues = [_get_button_value(button) for button in children]
    indices, set_indices = react.use_state([allvalues.index(k) for k in value], key="index")

    def on_indices(indices):
        set_indices(indices)
        value = [allvalues[k] for k in indices]
        if on_value:
            on_value(value)

    with v.BtnToggle(children=children, multiple=True, mandatory=False, v_model=indices, on_v_model=on_indices) as main:
        pass
    return main



from dataclasses import replace
from typing import List

import react_ipywidgets as react

import solara as sol
import solara.hooks.dataframe

from .. import CellAction, ColumnAction
from ..widgets import DataTable as DataTableWidget


def format_default(df, column, row_index, value):
    return str(value)


@react.component
def DataTable(df, page=0, items_per_page=20, format=None, column_actions: List[ColumnAction] = [], cell_actions: List[CellAction] = [], scrollable=False):
    total_length = len(df)
    options = {"descending": False, "page": page + 1, "itemsPerPage": items_per_page, "sortBy": [], "totalItems": total_length}
    options, set_options = react.use_state(options, key="options")
    format = format or format_default
    # frontend does 1 base, we use 0 based
    page = options["page"] - 1
    items_per_page = options["itemsPerPage"]
    i1 = page * items_per_page
    i2 = min(total_length, (page + 1) * items_per_page)

    columns = sol.hooks.dataframe.use_df_column_names(df)

    items = []
    column_data = {}
    dfs = df[i1:i2]

    if solara.hooks.dataframe.df_type(df) == "pandas":
        column_data = dfs[columns].to_dict("records")
    else:
        column_data = dfs[columns].to_records()
    for i in range(i2 - i1):
        item = {"__row__": i + i1}  # special key for the row number
        for column in columns:
            item[column] = format(dfs, column, i + i1, column_data[i][column])
        items.append(item)

    headers = [{"text": name, "value": name, "sortable": False} for name in columns]
    column_actions_callbacks = [k.on_click for k in column_actions]
    cell_actions_callbacks = [k.on_click for k in cell_actions]
    column_actions = [replace(k, on_click=None) for k in column_actions]
    cell_actions = [replace(k, on_click=None) for k in cell_actions]

    return DataTableWidget.element(
        total_length=total_length,
        items=items,
        headers=headers,
        headers_selections=[],
        options=options,
        items_per_page=items_per_page,
        selections=[],
        selection_colors=[],
        selection_enabled=False,
        highlighted=None,
        scrollable=scrollable,
        on_options=set_options,
        column_actions=column_actions,
        cell_actions=cell_actions,
        _column_actions_callbacks=column_actions_callbacks,
        _cell_actions_callbacks=cell_actions_callbacks,
    )



import textwrap
from typing import List

import pygments
import pymdownx.superfences
from pygments.formatters import HtmlFormatter
from pygments.lexers import get_lexer_by_name

from solara.kitchensink import react, sol


@react.component
def MarkdownIt(md_text: str, highlight: List[int] = []):
    # # from myst_parser.main import to_html

    # # html = to_html(md)
    # import markdown

    # # from ipywidgets import HTML

    # html = markdown.markdown(md, extensions=["codehilite"])
    # print(md, html)
    md_text = textwrap.dedent(md_text)

    from markdown_it import MarkdownIt as MarkdownItMod
    from mdit_py_plugins import container, deflist  # noqa: F401
    from mdit_py_plugins.footnote import footnote_plugin  # noqa: F401
    from mdit_py_plugins.front_matter import front_matter_plugin  # noqa: F401

    def highlight_code(code, name, attrs):
        """Highlight a block of code"""
        hl_lines = highlight
        if attrs:
            print(f"Ignoring {attrs}")

        if name:
            lexer = get_lexer_by_name(name)
            formatter = HtmlFormatter(hl_lines=hl_lines)  # linenos=True)
            return pygments.highlight(code, lexer, formatter)

    md = MarkdownItMod(
        "js-default",
        {
            # "linkify": True,
            "html": True,
            "typographer": True,
            "highlight": highlight_code,
        },
    )
    # tokens = md.parse(md)
    md = md.use(container.container_plugin, name="note")
    html = md.render(md_text)
    # print(html)

    return sol.HTML(unsafe_innerHTML=html)


extension_configs = {
    "pymdownx.superfences": {
        "custom_fences": [
            {
                "name": "mermaid",
                "class": "mermaid",
                "format": pymdownx.superfences.fence_div_format,
            },
        ],
    }
}


@react.component
def Markdown(md_text: str):
    import markdown

    md_text = textwrap.dedent(md_text)

    html = markdown.markdown(
        md_text,
        extensions=[
            "pymdownx.highlight",
            "pymdownx.superfences",
            "pymdownx.emoji",
            "toc",  # so we get anchors for h1 h2 etc
        ],
        extension_configs=extension_configs,
    )
    return sol.HTML(unsafe_innerHTML=html, class_="solara-markdown", style="max-width: 1024px;")



"""
Renders markdown using https://python-markdown.github.io/
"""


from solara.kitchensink import react, sol, v


@react.component
def MarkdownDemo():
    markdown_initial = """
# Large
## Smaller

## List items

    * item 1
    * item 2


## Code highlight support
```python
code = "formatted" and "supports highlighting"
```


## Mermaid support!
See [Mermaid docs](https://mermaid-js.github.io/)

```mermaid
graph TD;
    A-->B;
    A-->C;
    B-->D;
    C-->D;
```


    """.strip()
    markdown_text, set_markdown_text = react.use_state(markdown_initial)
    # with sol.GridFixed(columns=2) as main:
    with sol.HBox(grow=True) as main:
        with sol.VBox():
            sol.Markdown("# Input text")
            with sol.Padding(2):
                with v.Sheet(elevation=2):
                    v.Textarea(v_model=markdown_text, on_v_model=set_markdown_text, rows=30)
        with sol.VBox():
            sol.Markdown("# Renders like")
            with sol.Padding(2):
                with v.Sheet(elevation=2):
                    sol.Markdown(markdown_text)

    return main


Component = sol.Markdown
App = MarkdownDemo



import threading
import time
from typing import cast

import ipywidgets as widgets
import numpy as np
import react_ipywidgets as react
import react_ipywidgets.ipyvuetify as v
import react_ipywidgets.ipywidgets as w

from solara.components import MarkdownIt
from solara.hooks import use_thread

from .doc_use_download import DocUseDownload
from .docutils import IncludeComponent

x0 = np.linspace(0, 2, 100)


@react.component
def ButtonClick(label="Hi"):
    clicks, set_clicks = react.use_state(0)
    return w.Button(description=f"{label}: Clicked {clicks} times", on_click=lambda: set_clicks(clicks + 1))


@react.component
def Markdown(md: str):
    import markdown

    html = markdown.markdown(md)
    return w.HTML(value=html)


@react.component
def MarkdownEditor(md: str):
    md, set_md = react.use_state(md)
    with w.VBox() as main:
        w.Textarea(value=md, on_value=set_md)
        Markdown(md)
    return main


@react.component
def Doc():
    tab, set_tab = react.use_state(2, "tab")

    def set_meta():
        react.get_widget(main)._metadata = {"mount_id": "content2"}

    react.use_side_effect(set_meta)
    with v.Container() as main:
        with v.Alert(type="info", text=True, prominent=True, icon="mdi-school"):
            MarkdownIt("This documentation is live, and is running on Solara ☀️, find more on [Solara Github](https://github.com/widgetti/solara/)")
        with v.Tabs(v_model=tab, on_v_model=set_tab, vertical=True):
            for key in routes.keys():
                with v.Tab(children=[key]):
                    pass
            component = cast(react.core.Component, list(routes.values())[tab])
            with v.TabsItems(v_model=tab):
                component()
            #     pass
            # with v.Tab(children=["use_state"]):
            #     pass
            # with v.Tab(children=["use_effect"]):
            #     pass
            # with v.Tab(children=["use_memo"]):
            #     pass
            # with v.Tab(children=["use_reducer"]):
            #     pass
            # with v.Tab(children=["use_reducer"]):
            #     pass
            # with v.TabsItems(v_model=tab):
            #     if tab == 0:
            #         DocUseState()
            #     if tab == 1:
            #         DocUseEffect()
    return main


@react.component
def DocUseState(x=x0, ymax=5):
    with v.Container() as main:
        with w.VBox(layout={"padding": "20px", "max_width": "1024px"}):
            MarkdownIt(
                """
# use_state

```python
def use_state(initial: T, key: str = None) -> Tuple[T, Callable[[T], T]]:
    ...
```

use_state can be used to create a variable that is local to this component, and will be preserved during rerenders.

It returns a tuple with the current value, and a setter function that should be called to change the variable. A call to this setter
will trigger a rerender, and will cause the `use_state` function to return the new value on the next render.

## Simple examples

### Click button

Lets start with a Button, that renders how many times it is clicked.
        """
            )
            IncludeComponent(
                ButtonClick,
                """
import react_ipywidgets as react
import react_ipywidgets.ipywidgets as w

""",
                highlight=[6],
            )
            MarkdownIt(
                """
### Markdown editor
Lets continue with a more typical pattern, and create new new markdown component
        """
            )
            IncludeComponent(Markdown, md="# This is a custom\nMark-*down* **component**")

            MarkdownIt(
                """This component does not have state itself, the markdown text can only be set via its argument.
A common pattern is then to have its parent component manage the state, and pass it down:
"""
            )
            IncludeComponent(MarkdownEditor, md="# Edit me\nand the markdown component **will** *update*", highlight=[3, 5, 6])
            MarkdownIt(
                """Here we see the `MarkdownEditor` component using the `use_state` function to store the markdown text, while letting the `Textarea` component
                change its value"""
            )
    return main


@react.component
def CustomClickEvent():
    clicks, set_clicks = react.use_state(0)

    def attach_click_handler():
        def click_handler(*_ignore_args):
            set_clicks(clicks + 1)

        def cleanup():
            button_real.on_click(click_handler, remove=True)

        button_real: widgets.Button = react.get_widget(button)
        button_real.on_click(click_handler)
        return cleanup

    react.use_side_effect(attach_click_handler)
    button = w.Button(description=f"Clicked {clicks} times")
    return button


@react.component
def LongRunning():
    status, set_state = react.use_state("Doing nothing", key="status")
    progress, set_progress = react.use_state(0, key="progress")
    work_dependency, set_work_dependency = react.use_state(0, key="dependency")

    def do_work():
        cancel_event = threading.Event()

        def run():
            set_state("Running...")
            for i in range(101):
                time.sleep(0.1)
                set_progress(i)
                if cancel_event.is_set():
                    set_state("Cancelled")
                    break
            else:
                set_state("Done")

        def cleanup():
            cancel_event.set()
            try:
                worker_thread.join()
            except RuntimeError:
                pass  # fine, maybe it already stopped

        worker_thread = threading.Thread(target=run)
        worker_thread.start()
        return cleanup

    react.use_side_effect(do_work, [work_dependency])
    with w.VBox() as main:
        w.Label(value=status)
        w.FloatProgress(value=progress)
        w.Button(description="Restart", on_click=lambda: set_work_dependency(work_dependency + 1))
    return main


@react.component
def LongRunningEasier():
    status, set_state = react.use_state("Doing nothing", key="status")
    progress, set_progress = react.use_state(0, key="progress")
    work_dependency, set_work_dependency = react.use_state(0, key="dependency")

    def work(cancel_event: threading.Event):
        set_state("Running...")
        for i in range(101):
            time.sleep(0.1)
            set_progress(i)
            if cancel_event.is_set():
                set_state("Cancelled")
                break
        else:
            set_state("Done")

    use_thread(work, [work_dependency])
    with w.VBox() as main:
        w.Label(value=status)
        w.FloatProgress(value=progress)
        w.Button(description="Restart", on_click=lambda: set_work_dependency(work_dependency + 1))
    return main


@react.component
def DocUseEffect(x=x0, ymax=5):
    with v.Container() as main:  # () as main:
        with w.VBox(layout={"padding": "20px", "max_width": "1024px"}):
            MarkdownIt(
                """
# use_side_effect

```python

# types used:
EffectCleanupCallable = Callable[[], None]
EffectCallable = Callable[[], Optional[EffectCleanupCallable]]

def use_side_effect(effect: EffectCallable, dependencies=None):
    ...
```

use_side_effect can be used to run a function after the widgets are created, to optionally perform so called 'side effects', while at the same time providing a
function that will clean up any side effect. Typical side effects include attaching event handlers, fetching data and starting threads or tasks.

Providing a cleanup function ensure our app can always dispose of resources when needed. If no dependencies are passed, the function will be executed after each
render phase. Before your effect function is called for the next time, the cleanup function will be called.


## Simple examples

###

Lets create a Button again, like we did in `use_state`, but now we manually attach the click handler.
        """
            )
            IncludeComponent(
                CustomClickEvent,
                """
import ipywidgets as widgets

""",
                highlight=[13, 15],
            )
            MarkdownIt(
                """
Note that we return a `cleanup` function, so `react-ipywidgets` knows how to remove the event handler again. Note that under water, `on_click` does exactly
this for you, but with much less code. Note that we use `get_widget` to get a reference to the real ipywidgets `Button`
(this function can only be called from within an effect function).

### Long running jobs

If we want to run a long running job, we do not want to do that in the render loop, since the render loop can be executed multiple times. Also, doing work
in the render loop will cause our app to be unresponsive.

A better way would be to include the long running task in a thread, which will be started by the side effect hook, as well as a cleanup function that will
stop the execution of the work.

        """
            )
            IncludeComponent(
                LongRunning,
                """

import threading
import time

""",
                highlight=[8, 29],
            )
            MarkdownIt(
                """
Note that we implement a restart by adding an extra dependency (`work_dependency`) which when changed, will cancel the previous work, and start the new thread.


Since this pattern is so common, this funtionality is included in a user hook, called `use_thread`, making is much easier to include such long runnings tasks
in your components.
"""
            )
            IncludeComponent(
                LongRunningEasier,
                """
from solara.hooks.hooks import use_thread

""",
                highlight=[1, 20],
            )

    return main


app = Doc()


routes = {"use_state": DocUseState, "use_effect": DocUseEffect, "use_download": DocUseDownload}



from enum import Enum

import pydantic


class ThemeVariant(str, Enum):
    light = "light"
    dark = "dark"
    auto = "auto"


class ThemeSettings(pydantic.BaseSettings):
    variant: ThemeVariant = ThemeVariant.light
    variant_user_selectable: bool = True
    loader: str = "solara"

    class Config:
        env_prefix = "solara_theme_"
        case_sensitive = False
        env_file = ".env"


class MainSettings(pydantic.BaseSettings):
    use_pdb: bool = False
    mode: str = "production"

    class Config:
        env_prefix = "solara_"
        case_sensitive = False
        env_file = ".env"


main = MainSettings()
theme = ThemeSettings()



"""Build webapps using IPywidgets"""
__version__ = "0.4.1"
# isort: skip_file

from . import util

# flake8: noqa: F402
from .datatypes import *
from .hooks import *
from .memoize import memoize
from .components import *

github_url = "https://github.com/widgetti/solara"
git_branch = "master"



import ipycanvas
import numpy as np
import react_ipywidgets as react
from react_ipywidgets import ipycanvas as c
from react_ipywidgets import ipywidgets as w


def polygon(canvas, x, y, radius1, radius2, n_points):
    index = np.arange(n_points)
    angles = (2 * np.pi / n_points) * index

    radius = np.where(index % 2, radius1, radius2)
    v_x = x + np.cos(angles) * radius
    v_y = y + np.sin(angles) * radius

    points = np.stack((v_x, v_y), axis=1)

    canvas.fill_polygon(points)
    canvas.stroke_polygon(points)


@react.component
def CanvasDemo():
    width, height = 800, 800
    view_count, set_view_count = react.use_state(0)
    with w.ViewcountVBox(set_view_count) as main:
        fill = w.color("#63934e", "fill color")
        stroke = w.color("#4e6393", "stroke color")
        line_width = w.slider_int(4, description="line width", min=0, max=30)
        n_points = w.slider_int(5, "Points", min=1, max=8) * 2
        radius_inner = w.slider_float(30, "Inner radius", min=0, max=100)
        radius_outer = w.slider_float(80, "Outer radius", min=0, max=100)

        def real_drawing():
            canvas: ipycanvas.Canvas = react.core.get_widget(canvas_element)

            with ipycanvas.hold_canvas(canvas):
                canvas.clear()
                canvas.fill_style = fill
                canvas.stroke_style = stroke
                canvas.line_width = line_width
                radius = width // 3
                polygon(canvas, width // 2, height // 2, radius * radius_inner / 100, radius * radius_outer / 100, n_points)

        react.use_side_effect(real_drawing, [fill, stroke, line_width, n_points, view_count, radius_inner, radius_outer])
        canvas_element = c.Canvas(width=width, height=height)

    return main


app = CanvasDemo()



import numpy as np
import plotly.express as px

from solara.kitchensink import react, sol

x = np.linspace(0, 2, 100)


@react.component
def App():
    freq, set_freq = react.use_state(2.0)
    phase, set_phase = react.use_state(0.1)
    y = np.sin(x * freq + phase)

    with sol.VBox() as main:
        sol.FloatSlider("Frequency", value=freq, on_value=set_freq, min=0, max=10)
        sol.FloatSlider("Phase", value=phase, on_value=set_phase, min=0, max=np.pi, step=0.1)

        fig = px.line(x=x, y=y)
        sol.FigurePlotly(fig)
    return main


app = App()



import logging
import os
import typing
from typing import List, Optional, Union, cast

import anyio
import starlette.websockets
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse
from starlette.routing import Mount, Route, WebSocketRoute
from starlette.staticfiles import StaticFiles

from . import app as appmod
from . import patch, server, settings, websocket

logger = logging.getLogger("solara.server.fastapi")
# if we add these to the router, the server_test does not run (404's)
prefix = ""


class WebsocketWrapper(websocket.WebsocketWrapper):
    ws: starlette.websockets.WebSocket

    def __init__(self, ws: starlette.websockets.WebSocket, portal: anyio.from_thread.BlockingPortal) -> None:
        self.ws = ws
        self.portal = portal

    def close(self):
        self.portal.call(self.close)

    def send_text(self, data: str) -> None:
        self.portal.call(self.ws.send_text, data)

    def send_bytes(self, data: bytes) -> None:
        self.portal.call(self.ws.send_bytes, data)

    def receive(self):
        message = self.portal.call(self.ws.receive)
        if "text" in message:
            return message["text"]
        elif "bytes" in message:
            return message["bytes"]
        elif message.get("type") == "websocket.disconnect":
            raise websocket.WebSocketDisconnect()
        else:
            raise RuntimeError(f"Unknown message type {message}")


async def kernels(id):
    return JSONResponse({"name": "lala", "id": "dsa"})


async def kernel_connection(ws: starlette.websockets.WebSocket):
    context_id = ws.cookies.get(appmod.COOKIE_KEY_CONTEXT_ID)
    logger.info("Solara kernel requested for context_id %s", context_id)
    await ws.accept()

    def websocket_thread_runner(ws: starlette.websockets.WebSocket, context_id: str, portal: anyio.from_thread.BlockingPortal):
        ws_wrapper = WebsocketWrapper(ws, portal)

        async def run():
            try:
                await server.app_loop(ws_wrapper, context_id)
            except:  # noqa
                await portal.stop(cancel_remaining=True)
                raise

        anyio.run(run)

    # this portal allows us to sync call the websocket calls from this current event loop we are in
    # each websocket however, is handled from a separate thread
    try:
        async with anyio.from_thread.BlockingPortal() as portal:
            thread_return = anyio.to_thread.run_sync(websocket_thread_runner, ws, context_id, portal)
            await thread_return
    finally:
        try:
            await ws.close()
        except:  # noqa
            pass


async def watchdog(ws: starlette.websockets.WebSocket):
    await ws.accept()
    context_id = ws.cookies.get(appmod.COOKIE_KEY_CONTEXT_ID)

    def websocket_thread_runner(ws: starlette.websockets.WebSocket, context_id: str, portal: anyio.from_thread.BlockingPortal):
        ws_wrapper = WebsocketWrapper(ws, portal)

        async def run():
            try:
                server.control_loop(ws_wrapper, context_id)
            except:  # noqa
                await portal.stop(cancel_remaining=True)
                raise

        anyio.run(run)

    # this portal allows us to sync call the websocket calls from this current event loop we are in
    # each websocket however, is handled from a separate thread
    try:
        async with anyio.from_thread.BlockingPortal() as portal:
            thread_return = anyio.to_thread.run_sync(websocket_thread_runner, ws, context_id, portal)
            await thread_return
    finally:
        try:
            await ws.close()
        except:  # noqa
            pass


async def root(request: Request, fullpath: Optional[str] = ""):
    root_path = request.scope.get("root_path", "")
    logger.debug("root_path: %s", root_path)
    if request.headers.get("script-name"):
        logger.debug("override root_path using script-name header from %s to %s", root_path, request.headers.get("script-name"))
        root_path = request.headers.get("script-name")
    if request.headers.get("x-script-name"):
        logger.debug("override root_path using x-script-name header from %s to %s", root_path, request.headers.get("x-script-name"))
        root_path = request.headers.get("x-script-name")

    context_id = request.cookies.get(appmod.COOKIE_KEY_CONTEXT_ID)
    content, context_id = server.read_root(context_id, root_path)
    assert context_id is not None
    response = HTMLResponse(content=content)
    response.set_cookie(appmod.COOKIE_KEY_CONTEXT_ID, value=context_id)
    return response


class StaticNbFiles(StaticFiles):
    def get_directories(
        self, directory: Union[str, "os.PathLike[str]", None] = None, packages: typing.List[str] = None
    ) -> List[Union[str, "os.PathLike[str]"]]:
        return cast(List[Union[str, "os.PathLike[str]"]], server.nbextensions_directories)

    # allow us to follow symlinks, maybe only in dev mode?
    # from https://github.com/encode/starlette/pull/1377/files
    def lookup_path(self, path: str) -> typing.Tuple[str, typing.Optional[os.stat_result]]:
        if settings.main.mode == "production":
            return super().lookup_path(path)
        if settings.main.mode == "development":
            for directory in self.all_directories:
                original_path = os.path.join(directory, path)
                full_path = os.path.realpath(original_path)
                directory = os.path.realpath(directory)
                try:
                    return full_path, os.stat(full_path)
                except (FileNotFoundError, NotADirectoryError):
                    continue
            return "", None
        raise ValueError(f"Unknown mode {settings.main.mode}")


routes = [
    Route("/jupyter/api/kernels/{id}", endpoint=kernels),
    WebSocketRoute("/jupyter/api/kernels/{id}/{name}", endpoint=kernel_connection),
    WebSocketRoute("/solara/watchdog/", endpoint=watchdog),
    Route("/", endpoint=root),
    Route("/{fullpath}", endpoint=root),
    Mount(f"{prefix}/static/dist", app=StaticFiles(directory=server.voila_static)),
    Mount(f"{prefix}/static/nbextensions", app=StaticNbFiles()),
    Mount(f"{prefix}/static/nbconvert", app=StaticFiles(directory=server.nbconvert_static)),
    Mount(f"{prefix}/static", app=StaticFiles(directory=server.solara_static)),
]
app = Starlette(routes=routes)
patch.patch()



import json
import logging
import struct
from typing import Set

import ipykernel.kernelbase
import jupyter_client.session as session
from ipykernel.comm import CommManager
from jupyter_client.jsonutil import json_default
from jupyter_client.session import json_packer
from zmq.eventloop.zmqstream import ZMQStream

from . import websocket

# from notebook.base.zmqhandlers import serialize_binary_message
# this saves us a depdendency on notebook/jupyter_server when e.g.
# running on pyodide


def serialize_binary_message(msg):
    """serialize a message as a binary blob

    Header:

    4 bytes: number of msg parts (nbufs) as 32b int
    4 * nbufs bytes: offset for each buffer as integer as 32b int

    Offsets are from the start of the buffer, including the header.

    Returns
    -------

    The message serialized to bytes.

    """
    # don't modify msg or buffer list in-place
    msg = msg.copy()
    buffers = list(msg.pop("buffers"))
    bmsg = json.dumps(msg, default=json_default).encode("utf8")
    buffers.insert(0, bmsg)
    nbufs = len(buffers)
    offsets = [4 * (nbufs + 1)]
    for buf in buffers[:-1]:
        offsets.append(offsets[-1] + len(buf))
    offsets_buf = struct.pack("!" + "I" * (nbufs + 1), nbufs, *offsets)
    buffers.insert(0, offsets_buf)
    return b"".join(buffers)


SESSION_KEY = b"solara"


class WebsocketStream(object):
    def __init__(self, session, channel: str):
        self.session = session
        self.channel = channel


class BytesWrap(object):
    def __init__(self, bytes):
        self.bytes = bytes


class WebsocketStreamWrapper(ZMQStream):
    def __init__(self, websocket, channel):
        self.websocket = websocket
        self.channel = channel

    def flush(self, *ignore):
        pass


def send_websockets(websockets: Set[websocket.WebsocketWrapper], binary_msg):
    for ws in list(websockets):
        try:
            ws.send(binary_msg)
        except:  # noqa
            # in case of any issue, we simply remove it from the list
            websockets.remove(ws)


class SessionWebsocket(session.Session):
    def __init__(self, *args, **kwargs):
        super(SessionWebsocket, self).__init__(*args, **kwargs)
        self.websockets: Set[websocket.WebsocketWrapper] = set()  # map from .. msg id to websocket?

    def send(self, stream, msg_or_type, content=None, parent=None, ident=None, buffers=None, track=False, header=None, metadata=None):
        try:
            msg = self.msg(msg_or_type, content=content, parent=parent, header=header, metadata=metadata)
            msg["channel"] = stream.channel
            if buffers:
                msg["buffers"] = [memoryview(k).cast("b") for k in buffers]
                binary_msg = serialize_binary_message(msg)
            else:
                binary_msg = json_packer(msg).decode("utf8")
            send_websockets(self.websockets, binary_msg)
        except Exception as e:
            print(e)


class Kernel(ipykernel.kernelbase.Kernel):
    implementation = "ipython"
    implementation_version = "0.1"
    banner = "banner"

    def __init__(self):
        super(Kernel, self).__init__()
        self.session = SessionWebsocket(parent=self, key=SESSION_KEY)

        self.stream = self.iopub_socket = WebsocketStream(self.session, "iopub")
        # on github action the next line gives a mypy error:
        # solara/server/kernel.py:111: error: "SessionWebsocket" has no attribute "stream"
        # not sure why we cannot reproduce that locally
        self.session.stream = self.iopub_socket  # type: ignore
        self.comm_manager = CommManager(parent=self, kernel=self)
        self.shell = None
        self.log = logging.getLogger("fake")

        comm_msg_types = ["comm_open", "comm_msg", "comm_close"]
        for msg_type in comm_msg_types:
            self.shell_handlers[msg_type] = getattr(self.comm_manager, msg_type)

    async def _flush_control_queue(self):
        pass

    # these don't work from non-main thread, and we do not care about them I think
    # TODO: it seems that if post_handler_hook is not override, the flask reload tests fails
    # for unknown reason
    def pre_handler_hook(self, *args):
        pass

    def post_handler_hook(self, *args):
        pass



import os
from datetime import date, datetime, timedelta
from typing import Callable, List, TypeVar

import ipyvue
import traitlets

from solara.kitchensink import react, v

T = TypeVar("T")


@react.component
def FloatSlider(label: str, value: float = 0, min: float = 0, max: float = 10.0, step: float = 0.1, on_value: Callable[[float], None] = None, thumb_label=True):
    def set_value_cast(value):
        if on_value is None:
            return
        on_value(float(value))

    return v.Slider(v_model=value, on_v_model=set_value_cast, label=label, min=min, max=max, step=step, thumb_label=thumb_label, dense=False, hide_details=True)


@react.component
def IntSlider(label: str, value: int = 0, min: int = 0, max: int = 10, step: int = 1, on_value: Callable[[int], None] = None, thumb_label=True):
    def set_value_cast(value):
        if on_value is None:
            return
        on_value(int(value))

    return v.Slider(v_model=value, on_v_model=set_value_cast, label=label, min=min, max=max, step=step, thumb_label=thumb_label, dense=False, hide_details=True)


@react.component
def ValueSlider(label: str, value: T, values: List[T], on_value: Callable[[T], None] = None):
    index, set_index = react.use_state(values.index(value), key="index")

    def on_index(index):
        set_index(index)
        value = values[index]
        print(index, value)
        if on_value:
            on_value(value)

    return v.Slider(
        v_model=index,
        on_v_model=on_index,
        ticks=True,
        tick_labels=values,
        label=label,
        min=0,
        max=len(values) - 1,
        dense=False,
        hide_details=True,
    )


class DateSliderWidget(ipyvue.VueTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "slider_date.vue"))

    min = traitlets.CFloat(0).tag(sync=True)
    days = traitlets.CFloat(0).tag(sync=True)
    value = traitlets.Any(0).tag(sync=True)

    label = traitlets.Unicode("").tag(sync=True)


@react.component
def DateSlider(
    label: str, value: date = date(1981, 7, 28), min: date = date(1950, 1, 1), max: date = date(3000, 12, 30), on_value: Callable[[date], None] = None
):
    def format(d: date):
        return float(datetime(d.year, d.month, d.day).timestamp())

    dt_min = format(min)
    delta: timedelta = max - min
    days = delta.days

    delta_value: timedelta = value - min
    days_value = delta_value.days
    if days_value < 0:
        days_value = 0

    def set_value_cast(value):
        date = min + timedelta(days=value)
        if on_value:
            on_value(date)

    return DateSliderWidget.element(label=label, min=dt_min, days=days, on_value=set_value_cast, value=days_value)



import os
import shutil
import site
import sys
import threading
import time
import typing
import webbrowser
from enum import Enum
from pathlib import Path

import react_ipywidgets
import rich
import rich_click as click
import uvicorn
from uvicorn.main import LEVEL_CHOICES

import solara

from .server import settings

HERE = Path(__file__).parent
HOST_DEFAULT = os.environ.get("HOST", "localhost")
if "arm64-apple-darwin" in HOST_DEFAULT:  # conda activate script
    HOST_DEFAULT = "localhost"

LOGGING_CONFIG: dict = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "fmt": "%(levelprefix)s %(message)s",
            "use_colors": None,
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "fmt": '%(levelprefix)s %(client_addr)s - "%(request_line)s" %(status_code)s',  # noqa: E501
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "rich": {
            "class": "rich.logging.RichHandler",
        },
        "access": {
            "formatter": "access",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "solara": {"handlers": ["rich"], "level": "INFO"},
        # "react": {"handlers": ["rich"], "level": "DEBUG"},
        "uvicorn": {"handlers": ["default"], "level": "ERROR"},
        "uvicorn.error": {"level": "ERROR"},
        "uvicorn.access": {"handlers": ["access"], "level": "ERROR", "propagate": False},
    },
}


def find_all_packages_paths():
    paths = []
    # sitepackages = set([os.path.dirname(k) for k in site.getsitepackages()])
    sitepackages = set([k for k in site.getsitepackages()])
    paths.extend(list(sitepackages))
    print(sitepackages)
    for name, module in sys.modules.items():
        if hasattr(module, "__path__"):
            try:
                path = module.__path__[0]
            except:  # noqa: E722
                pass  # happens for namespace packages it seems
                # print(f"Error for {name}")
                # if path:
                #     skip = False
                #     for sitepackage in sitepackages:
                #         if path.startswith(sitepackage):
                #             skip = True
                # if not skip:
                # print(name, path, skip)
                paths.append(str(path))
    # print("PATHS", paths)
    return paths


class EnumType(click.Choice):
    def __init__(self, enum: typing.Type[Enum], case_sensitive=False):
        self._enum = enum
        super().__init__(choices=[item.name for item in enum], case_sensitive=case_sensitive)


@click.group()
def cli():
    pass


@cli.command()
@click.option("--port", default=int(os.environ.get("PORT", 8765)))
@click.option("--host", default=HOST_DEFAULT)
@click.option(
    "--dev/--no-dev",
    default=False,
    help="""Tell Solara to work in production(default) or development mode.
When in dev mode Solara will:
  Auto reload server when the server code changes
  Prefer non-minized js/css assets for easier debugging.
""",
)
@click.option("--open/--no-open", default=False)
@click.option("--reload", is_flag=True, default=False, help="Enable auto-reload.")
@click.option(
    "--reload-dir",
    "reload_dirs",
    multiple=True,
    help="Set reload directories explicitly, instead of using the current working" " directory.",
    type=click.Path(exists=True),
)
@click.option(
    "--reload-exclude",
    "reload_excludes",
    multiple=True,
    help="Set glob patterns to exclude while watching for files. Includes "
    "'.*, .py[cod], .sw.*, ~*' by default; these defaults can be overridden "
    "with `--reload-include`. This option has no effect unless watchgod is "
    "installed.",
)
@click.option(
    "--workers",
    default=None,
    type=int,
    help="Number of worker processes. Defaults to the $WEB_CONCURRENCY environment" " variable if available, or 1. Not valid with --reload.",
)
@click.option(
    "--env-file",
    type=click.Path(exists=True),
    default=None,
    help="Environment configuration file.",
    show_default=True,
)
@click.option(
    "--log-config",
    type=click.Path(exists=True),
    default=None,
    help="Logging configuration file. Supported formats: .ini, .json, .yaml.",
    show_default=True,
)
@click.option(
    "--log-level",
    type=LEVEL_CHOICES,
    default=None,
    help="Log level. [default: info]",
    show_default=True,
)
@click.option(
    "--log-level-uvicorn",
    type=LEVEL_CHOICES,
    default="error",
    help="Log level. [default: error]",
    show_default=True,
)
@click.option(
    "--access-log/--no-access-log",
    is_flag=True,
    default=True,
    help="Enable/Disable access log.",
)
@click.option(
    "--root-path",
    type=str,
    default="",
    help="Set the ASGI 'root_path' for applications submounted below a given URL path.",
)
@click.option(
    "--theme-loader",
    type=str,
    default=settings.theme.loader,
    help=f"Loader to use when the app is not yet shown to the user. [default: {settings.theme.loader!r}]",
)
@click.option(
    "--theme-variant",
    type=settings.ThemeVariant,
    default=settings.ThemeVariant.light.name,
    help=f"Use light or dark variant, or auto detect (auto). [default: {settings.theme.variant.name}",
)
@click.option(
    "--theme-variant-user-selectable/--no-theme-variant-user-selectable",
    type=bool,
    default=settings.theme.variant_user_selectable,
    help=f"Can the user select the theme variant from the UI. [default: {settings.theme.variant_user_selectable}",
)
@click.option("--pdb/--no-pdb", "use_pdb", default=False, help="Enter debugger on error")
@click.argument("app")
def run(
    app,
    host,
    port,
    open,
    reload: bool,
    reload_dirs: typing.Optional[typing.List[str]],
    dev: bool,
    reload_excludes: typing.List[str],
    workers: int,
    env_file: str,
    root_path: str,
    log_config: str,
    log_level: str,
    log_level_uvicorn: str,
    access_log: bool,
    use_pdb: bool,
    theme_loader: str,
    theme_variant: settings.ThemeVariant,
    theme_variant_user_selectable: bool,
):
    reload_dirs = reload_dirs if reload_dirs else None
    url = f"http://{host}:{port}"

    failed = False
    if dev:
        solara_root = Path(solara.__file__).parent

        reload_dirs = list(reload_dirs if reload_dirs else [])

        # we restart the server when solara or react changes, in priciple we should do
        # that for all dependencies of the server, but these are changing most often
        # during development
        # We exclude exampes, that will be handled by solara/server/reload.py
        reload_dirs = [str(solara_root), str(Path(react_ipywidgets.__file__).parent)]
        reload_excludes = reload_excludes if reload_excludes else []
        reload_excludes = [str(solara_root / "examples")]
        del solara_root
        reload = True
        settings.main.mode = "development"

    server = None

    # TODO: we might want to support this, but it needs to be started from the main thread
    # and then uvicorn needs to be started from a thread
    # def open_webview():
    #     import webview

    #     while not failed and (server is None or not server.started):
    #         time.sleep(0.1)
    #     if not failed:
    #         window = webview.create_window("Hello world", url, resizable=True)
    #         window.on_top = True
    #         # window.show()
    #         webview.start(debug=True)

    def open_browser():
        while not failed and (server is None or not server.started):
            time.sleep(0.1)
        if not failed:
            webbrowser.open(url)

    if open:
        threading.Thread(target=open_browser, daemon=True).start()
    rich.print(f"Solara server is starting at {url}")

    if log_level is not None:
        LOGGING_CONFIG["loggers"]["solara"]["level"] = log_level.upper()

    log_level = log_level_uvicorn
    del log_level_uvicorn

    kwargs = locals().copy()
    # cgi vars: https://datatracker.ietf.org/doc/html/rfc3875
    os.environ["SOLARA_APP"] = app
    os.environ["SERVER_NAME"] = host
    os.environ["SERVER_PORT"] = str(port)

    kwargs["app"] = "solara.server.starlette:app"
    kwargs["log_config"] = LOGGING_CONFIG if log_config is None else log_config
    settings.main.use_pdb = use_pdb
    settings.theme.loader = theme_loader
    settings.theme.variant = theme_variant
    settings.theme.variant_user_selectable = theme_variant_user_selectable
    for item in "theme_variant_user_selectable theme_variant theme_loader use_pdb server open_browser open url failed dev".split():
        del kwargs[item]

    def start_server():
        nonlocal server
        nonlocal failed
        try:
            # we manually create the server instead of calling uvicorn.run
            # because we can then access the server variable and check if it is
            # running.
            config = uvicorn.Config(**kwargs)
            server = uvicorn.Server(config=config)
            if reload:
                sock = config.bind_socket()
                from uvicorn.supervisors import ChangeReload

                ChangeReload(config, target=run_with_settings(server, main=settings.main.dict(), theme=settings.theme.dict()), sockets=[sock]).run()
            else:
                server.run()
        except:  # noqa
            failed = True
            raise

    start_server()

    # TODO: if we want to use webview, it should be sth like this
    # server_thread = threading.Thread(target=start_server)
    # server_thread.start()
    # if open:
    #     # open_webview()
    #     open_browser()
    # server_thread.join()


class run_with_settings:
    """This cross a process boundry, and takes the serialized settings with it"""

    def __init__(self, server: uvicorn.Server, main: typing.Dict, theme: typing.Dict):
        self.server = server
        self.main = main
        self.theme = theme

    def __call__(self, *args, **kwargs):
        # this is now in the new process, where we need to re-apply the settings
        settings.main = settings.MainSettings(**self.main)
        settings.theme = settings.ThemeSettings(**self.theme)
        return self.server.run(*args, **kwargs)


@cli.command()
def staticbuild():
    """Experimental static build"""
    # imports locals, otherwise .run() does not use the cli arguments
    import solara.server
    import solara.server.patch
    import solara.server.server

    server_path = HERE / "server/static"

    build_dir = Path("staticbuild")
    build_dir.mkdir(exist_ok=True)

    build_dir_static = build_dir / "static"
    build_dir_static.mkdir(exist_ok=True)
    for path in (
        list(server_path.glob("*.css"))
        + list(server_path.glob("*.js"))
        + list(server_path.glob("*.ps"))
        + list(server_path.glob("*.py"))
        + list(server_path.glob("*.png"))
        + list(server_path.glob("*.svg"))
    ):
        shutil.copy(path, build_dir_static)

    include_nbextensions = True
    if include_nbextensions:
        directories = solara.server.server.get_nbextensions_directories()
        nbextensions = solara.server.server.get_nbextensions()
        for name in nbextensions:
            for directory in directories:
                if (directory / (name + ".js")).exists():
                    src = (directory / (name + ".js")).parent
                    dst: Path = (build_dir_static / "nbextensions" / (name + ".js")).parent
                    dst.mkdir(parents=True, exist_ok=True)
                    shutil.copytree(src, dst)

    build_dir_wheels = build_dir / "wheels"
    build_dir_wheels.mkdir(exist_ok=True)
    version = solara.__version__
    for path in [Path(f"dist/solara-{version}-py2.py3-none-any.whl")]:
        shutil.copy(path, build_dir_wheels)

    build_dir_nbconvert_static = build_dir / "static/nbconvert"
    build_dir_nbconvert_static.mkdir(exist_ok=True, parents=True)
    nbconvert = Path(solara.server.server.nbconvert_static)
    for path in list(nbconvert.glob("*.js")) + list(nbconvert.glob("*.css")):
        shutil.copy(path, build_dir_nbconvert_static)

    build_dir_static_dist = build_dir / "static/dist"
    build_dir_static_dist.mkdir(exist_ok=True)
    voila = server_path / "dist"
    for path in list(voila.glob("*.js")) + list(voila.glob("*.woff")):
        shutil.copy(path, build_dir_static_dist)

    solara.server.patch.patch()
    index_html, context_id = solara.server.server.read_root("fake-context-id", "", render_kwargs={"for_pyodide": True}, use_nbextensions=True)
    (build_dir / "index.html").write_text(index_html)


cli.add_command(run)
cli.add_command(staticbuild)


def main():
    cli()


if __name__ == "__main__":
    main()



from solara.kitchensink import react, v

from .calculator import Calculator
from .pokemon import App as Pokemon

tabs = {
    "Calulator": Calculator,
    "Pokemon": Pokemon,
}


@react.component
def Demo():
    tab, set_tab = react.use_state(0, "tab")

    # md, set_md = use_state("")
    with v.Tabs(v_model=tab, on_v_model=set_tab, vertical=True) as main:
        for key in tabs:
            with v.Tab(children=[key]):
                pass
        component = list(tabs.values())[tab]
        with v.TabsItems(v_model=tab):
            component()
    return main



"""
# ToggleButtons

ToggleButtons are in two flavours, for single, and for multiple selections.


"""
from solara.kitchensink import react, sol


@react.component
def Test():
    with sol.VBox() as main:
        with sol.Card("Single selection"):
            food, set_food = react.use_state("banana")
            sol.Markdown(f"**Selected**: {food}")
            with sol.ToggleButtonsSingle(food, on_value=set_food):
                sol.Button("Kiwi")
                sol.Button("Banana", value="banana")  # override the default value (the button label)
                sol.Button("Apple")

        with sol.Card("Multiple selections"):
            all_languages = "Python C++ Java JavaScript TypeScript BASIC".split()
            languages, set_languages = react.use_state([all_languages[0]])
            sol.Markdown(f"**Selected**: {languages}")
            with sol.ToggleButtonsMultiple(languages, on_value=set_languages):
                for language in all_languages:
                    sol.Button(language, value=language)

    return main


App = Test
app = App()



# isort: skip_file
from .misc import *  # noqa: #F401 F403
from .datatable import DataTable  # noqa: #F401 F403
from .details import Details  # noqa: #F401 F403
from .file_browser import FileBrowser  # noqa: #F401 F403
from .image import Image  # noqa: #F401 F403
from .markdown import Markdown, MarkdownIt  # noqa: #F401 F403
from .slider import DateSlider, FloatSlider, IntSlider, ValueSlider  # noqa: #F401 F403
from .sql_code import SqlCode  # noqa: #F401 F403
from .togglebuttons import (  # noqa: #F401 F403
    ToggleButtonsMultiple,
    ToggleButtonsSingle,
)



"""
# DataTable

The DataTable component can render dataframes of any size due to pagination.

## API

### Component signature
```python
@react.component
def DataTable(df, page=0, items_per_page=20, format=None, column_actions: List[ColumnAction] = [], cell_actions: List[CellAction] = []):
    ...
```

### arguments

* `df` - `DataFrame`

### events

* `column_actions` - Triggered via clicking on the triple dot icon on the headers (visible when hovering).
* `cell_actions` -  Triggered via clicking on the triple dot icon in the cell (visible when hovering).

"""

from typing import Any, Dict, Optional, cast

try:
    import vaex
except ImportError:
    vaex = None
import solara.components.datatable
from solara.kitchensink import react, sol

if vaex is not None:
    df = vaex.datasets.titanic()
else:
    df = None


@react.component
def DataTableDemo():
    column, set_column = react.use_state(cast(Optional[str], None))
    cell, set_cell = react.use_state(cast(Dict[str, Any], {}))

    def on_action_column(column):
        set_column(column)

    def on_action_cell(column, row_index):
        set_cell(dict(column=column, row_index=row_index))

    column_actions = [sol.ColumnAction(icon="mdi-sunglasses", name="User column action", on_click=on_action_column)]
    cell_actions = [sol.CellAction(icon="mdi-white-balance-sunny", name="User cell action", on_click=on_action_cell)]
    with sol.Div() as main:
        sol.MarkdownIt(
            f"""
            ## Demo

            Below we show display the titanic dataset, and demonstrate a user colum and cell action. Try clicking on the triple icon when hovering
            above a column or cell. And see the following values changes:

            * Column action on: `{column}`
            * Cell action on: `{cell}`

        """
        )
        sol.components.datatable.DataTable(df, column_actions=column_actions, cell_actions=cell_actions)
    return main


Component = solara.components.datatable.DataTable
App = DataTableDemo



from typing import Callable, List

import ipyvue as vue
import react_ipywidgets as react
import react_ipywidgets.ipyvuetify as v
import react_ipywidgets.ipyvuetify as ipyvue

import solara as sol
import solara.widgets

PivotTable = react.core.ComponentWidget(solara.widgets.PivotTable)
Navigator = react.core.ComponentWidget(solara.widgets.Navigator)
GridDraggable = react.core.ComponentWidget(solara.widgets.GridLayout)
# keep the old name for a while
GridLayout = GridDraggable


@react.component
def ListItem(title, icon_name: str = None, children=[]):
    if children:
        with v.ListItem(value=title) as main:
            v.ListItemTitle(children=[title])
            with v.ListItemIcon():
                v.Icon(children=[icon_name or ""])
        return v.ListGroup(children=children, v_slots=[{"name": "activator", "children": main}], sub_group=True, value=True, append_icon=icon_name)
    else:
        with v.ListItem(value=title) as main:
            with v.ListItemIcon():
                v.Icon(children=[icon_name or ""])
            v.ListItemTitle(children=[title])
        return main


def ui_dropdown(label, value=None, options=["foo", "bar"], key=None, disabled=False, **kwargs):
    key = key or str(value) + str(label) + str(options)
    value, set_value = react.use_state(value, key)

    def set_index(index):
        set_value(options[index])

    v.Select(v_model=value, label=label, items=options, on_v_model=set_value, clearable=True, disabled=disabled, **kwargs)
    return value


def ui_text(label, value="", key=None, clearable=False, hint="", disabled=False, **kwargs):
    key = key or str(value) + str(label) + str(hint)
    value, set_value = react.use_state(value, key)
    v.TextField(v_model=value, label=label, on_v_model=set_value, clearable=clearable, hint=hint, disabled=disabled, **kwargs)
    return value


def ui_checkbox(label, value=True, key=None, disabled=False, **kwargs):
    key = key or str(value) + str(label)
    value, set_value = react.use_state(value, key)
    v.Checkbox(v_model=value, label=label, on_v_model=set_value, **kwargs)
    return value


def ui_slider(value=1, label="", min=0, max=100, key=None, tick_labels=None, thumb_label=None, disabled=False, **kwargs):
    key = key or str(value) + str(label)
    value, set_value = react.use_state(value, key)
    v.Slider(
        v_model=value,
        label=label,
        min=min,
        max=max,
        on_v_model=set_value,
        ticks=tick_labels is not None,
        tick_labels=tick_labels,
        thumb_label=thumb_label,
        disabled=disabled,
        **kwargs,
    )
    return value


@react.component
def Card(title: str = None, elevation: int = 2, margin=2, children: List[react.core.Element] = []):
    with v.Card(elevation=elevation, class_=f"ma-{margin}") as main:
        if title:
            with v.CardTitle(
                children=[title],
            ):
                pass
        with v.CardText(children=children):
            pass
    return main


@react.component
def Text(text):
    return vue.Html.element(tag="span", children=[text])


@react.component
def Div(children=[], **kwargs):
    return vue.Html.element(tag="div", children=children, **kwargs)


@react.component
def Preformatted(text, **kwargs):
    return vue.Html.element(tag="pre", children=[text], **kwargs)


@react.component
def Warning(text, icon="mdi-alert", children=[], **kwargs):
    return v.Alert(type="warning", text=True, prominent=True, icon=icon, children=[text, *children], **kwargs)


@react.component
def Info(text, icon="mdi-alert", children=[], progress=False, **kwargs):
    return v.Alert(type="info", text=True, prominent=True, icon=icon, children=[text, *children], **kwargs)


@react.component
def Error(text, icon="mdi-alert", children=[], **kwargs):
    return v.Alert(type="error", text=True, prominent=True, icon=icon, children=[text, *children], **kwargs)


@react.component
def Success(text, icon="mdi-alert", children=[], **kwargs):
    return v.Alert(type="success", text=True, prominent=True, icon=icon, children=[text, *children], **kwargs)


@react.component
def Button(
    label: str = None,
    on_click: Callable[[], None] = None,
    icon_name: str = None,
    children: list = [],
    click_event="click",
    disabled=False,
    value=None,
    **kwargs,
):
    if label:
        children = [label] + children
    if icon_name:
        children = [v.Icon(left=bool(label), children=[icon_name])] + children
    btn = v.Btn(children=children, **kwargs, disabled=disabled)
    ipyvue.use_event(btn, click_event, lambda *_ignore: on_click and on_click())
    return btn


@react.component
def IconButton(icon_name: str = None, on_click=Callable[[], None], children: list = [], click_event="click", **kwargs):
    return Button(icon_name=icon_name, on_click=on_click, children=children, icon=True, click_event=click_event, **kwargs)


@react.component
def HTML(tag="div", unsafe_innerHTML=None, style: str = None, class_: str = None, **kwargs):
    attributes = dict(style=style)
    attributes["class"] = class_
    return solara.widgets.HTML.element(tag=tag, unsafe_innerHTML=unsafe_innerHTML, attributes=attributes)


@react.component
def VBox(children=[], grow=True, align_items="stretch"):
    style = f"flex-direction: column; align-items: {align_items};"
    if grow:
        style += "flex-grow: 1;"
    return v.Sheet(class_="d-flex", style_=style, elevation=0, children=children)


@react.component
def HBox(children=[], grow=True, align_items="stretch"):
    style = f"flex-direction: row; align-items: {align_items}; "
    if grow:
        style += "flex-grow: 1;"
    return v.Sheet(class_="d-flex", style_=style, elevation=0, children=children)


@react.component
def GridFixed(columns=4, column_gap="10px", row_gap="10px", children=[], align_items="stretch", justify_items="stretch"):
    """

    See css grid spec:
    https://css-tricks.com/snippets/css/complete-guide-grid/
    """
    style = (
        f"display: grid; grid-template-columns: repeat({columns}, minmax(0, 1fr)); "
        + f"grid-column-gap: {column_gap}; grid-row-gap: {row_gap}; align-items: {align_items}; justify-items: {justify_items}"
    )
    return Div(style_=style, children=children)


@react.component
def Padding(size, children=[], grow=True):
    style = "flex-direction: row;"
    if grow:
        style += "flex-grow: 1;"
    return v.Sheet(class_=f"pa-{size}", style_=style, elevation=0, children=children)


@react.component
def AltairChart(chart, on_click=None, on_hover=None):
    import altair as alt

    with alt.renderers.enable("mimetype"):
        bundle = chart._repr_mimebundle_()[0]
        key = "application/vnd.vegalite.v4+json"
        if key not in bundle:
            raise KeyError(f"{key} not in mimebundle:\n\n{bundle}")
        spec = bundle[key]
        return solara.widgets.VegaLite.element(
            spec=spec, on_click=on_click, listen_to_click=on_click is not None, on_hover=on_hover, listen_to_hover=on_hover is not None
        )


@react.component
def FigurePlotly(fig, on_selection=None, on_click=None, on_hover=None, dependencies=None):
    from plotly.graph_objs._figurewidget import FigureWidget

    fig_element = FigureWidget.element()

    def update_data():
        fig_widget: FigureWidget = react.get_widget(fig_element)
        if on_selection:
            fig_widget.on_selection(on_selection)
        fig_widget.layout = fig.layout
        length = len(fig_widget.data)
        fig_widget.add_traces(fig.data)
        data = list(fig_widget.data)
        fig_widget.data = data[length:]
        for trace in fig_widget.data:
            if on_click:
                trace.on_click(on_click)
            if on_hover:
                trace.on_hover(on_hover)

    react.use_side_effect(update_data, dependencies or fig)
    return fig_element


@react.component
def Code(path, path_header=None):
    path_header = path_header or path
    with open(path) as f:
        code = f.read()
    md = sol.Markdown(
        f"""
### {path_header}

```python
{code}
```

"""
    )

    with v.ExpansionPanels() as main:
        with v.ExpansionPanel():
            with v.ExpansionPanelHeader(children=["View source"]):
                pass
            with v.ExpansionPanelContent(children=[md]):
                pass
    return main



"""
# VBox

Lays out children in a vertical direction.
"""

from solara.kitchensink import react, sol

from .common import ColorCard


@react.component
def VBoxDemo():
    with sol.VBox() as main:
        colors = "green red orange brown yellow pink".split()
        with sol.VBox():
            for color in colors:
                ColorCard(color, color)
    return main


Component = sol.VBox
App = VBoxDemo



from solara.hooks.misc import use_fetch
from solara.kitchensink import react, sol, v

github_url = sol.util.github_url(__file__)

json_url = "https://jherr-pokemon.s3.us-west-1.amazonaws.com/index.json"


@react.component
def App():
    data = use_fetch(json_url)
    json = sol.use_json_load(data)

    with sol.Div() as main:
        if json.error:
            v.Alert(children=[f"Error {json.error}"])
            sol.Button("Retry", on_click=data.retry)
        else:
            if json.value:
                filter = sol.ui_text(label="Filter pokemons by name", value="")
                pokemons = json.value
                if filter:
                    pokemons = [k for k in pokemons if filter.lower() in k["name"].lower()]
                    v.Label(children=[f"{len(pokemons)} pokemons found"])
                else:
                    v.Label(children=[f"{len(pokemons)} pokemons in total"])
                with sol.GridFixed(columns=4, align_items="end", justify_items="stretch"):
                    for pokemon in pokemons[:20]:
                        with sol.Div():
                            name = pokemon["name"]
                            url = "https://jherr-pokemon.s3.us-west-1.amazonaws.com/" + pokemon["image"]
                            v.Img(src=url, contain=True, max_height="200px")
                            v.Label(children=[name])
            else:
                with v.Text(children=["Loading..."]):
                    v.ProgressCircular(indeterminate=True, class_="solara-progress")
    return main


app = App()



import contextlib
import logging
import os
import sys
import threading
import time
from http.server import HTTPServer
from typing import Optional, Union

import pytest
import uvicorn.server

import solara.server.app
import solara.server.server
from solara.server import reload
from solara.server.starlette import app as app_starlette

logger = logging.getLogger("solara-test.integration")

worker = os.environ.get("PYTEST_XDIST_WORKER", "gw0")
TEST_PORT = 18765 + 2 + int(worker[2:])
SERVER = os.environ.get("SOLARA_SERVER")
if SERVER:
    SERVERS = [SERVER]
else:
    SERVERS = ["flask", "starlette"]


# see https://github.com/microsoft/playwright-pytest/issues/23
@pytest.fixture
def context(context):
    context.set_default_timeout(3000)
    yield context


class ServerBase(threading.Thread):
    def __init__(self, port: int, host: str = "localhost", **kwargs):
        self.port = port
        self.host = host
        self.base_url = f"http://{self.host}:{self.port}"

        self.kwargs = kwargs
        self.started = threading.Event()
        self.stopped = threading.Event()
        self.error: Optional[BaseException] = None
        self.server = None
        super().__init__(name="test-server-thread")
        self.setDaemon(True)

    def run(self):
        try:
            logger.info("Starting main loop")
            self.mainloop()
        except BaseException as e:  # noqa
            self.error = e
            self.started.set()
            logger.exception("Issue starting server")

    def serve_threaded(self):
        logger.debug("start thread")
        self.start()
        logger.debug("wait for thread to run")
        self.started.wait()
        if self.error:
            raise self.error

    def wait_until_serving(self):
        for n in range(30):
            if self.has_started():
                time.sleep(0.1)  # give some time to really start
                return
            time.sleep(0.05)
        else:
            raise RuntimeError(f"Server at {self.base_url} does not seem to be running")

    def serve(self):
        raise NotImplementedError

    def mainloop(self):
        logger.info("serving at http://%s:%d" % (self.host, self.port))
        try:
            self.serve()
        except:  # noqa: E722
            logger.exception("Oops, server stopped unexpectedly")
        finally:
            self.stopped.set()

    def stop_serving(self):
        logger.debug("stopping server")
        self.signal_stop()
        self.stopped.wait(10)
        if not self.stopped.is_set():
            logger.error("stopping server failed")
        else:
            logger.debug("stopped server")

    def signal_stop(self):
        pass

    def has_started(self) -> bool:
        return False


class ServerStarlette(ServerBase):
    server: uvicorn.server.Server

    def has_started(self):
        return self.server.started

    def signal_stop(self):
        self.server.should_exit = True
        # this cause uvicorn to not wait for background tasks, e.g.:
        # <Task pending name='Task-55'
        #  coro=<WebSocketProtocol.run_asgi() running at
        #  /.../uvicorn/protocols/websockets/websockets_impl.py:184>
        # wait_for=<Future pending cb=[<TaskWakeupMethWrapper object at 0x16896aa00>()]>
        # cb=[WebSocketProtocol.on_task_complete()]>
        self.server.force_exit = True
        self.server.lifespan.should_exit = True

    def serve(self):

        from uvicorn.config import Config
        from uvicorn.server import Server

        if sys.version_info[:2] < (3, 7):
            # make python 3.6 work
            import asyncio

            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # uvloop will trigger a: RuntimeError: There is no current event loop in thread 'fastapi-thread'
        config = Config(app_starlette, host=self.host, port=self.port, **self.kwargs, loop="asyncio")
        self.server = Server(config=config)
        self.started.set()
        self.server.run()


def _serve_flask_in_process(port, host):
    from solara.server.flask import app

    app.run(debug=False, port=port, host=host)


class ServerFlask(ServerBase):
    def has_started(self):
        import socket

        s = socket.socket()
        try:
            s.connect((self.host, self.port))
        except ConnectionRefusedError:
            return False
        return True

    def signal_stop(self):
        assert isinstance(self.server, HTTPServer)
        self.server.shutdown()  # type: ignore

    def serve(self):
        from werkzeug.serving import make_server

        from solara.server.flask import app

        self.server = make_server(self.host, self.port, app, threaded=True)  # type: ignore
        assert isinstance(self.server, HTTPServer)
        self.started.set()
        self.server.serve_forever(poll_interval=0.05)  # type: ignore


server_classes = {
    "flask": ServerFlask,
    "starlette": ServerStarlette,
}


@pytest.fixture(params=SERVERS)
def solara_server(request):
    server_class = server_classes[request.param]
    global TEST_PORT
    webserver = server_class(TEST_PORT)

    try:
        webserver.serve_threaded()
        webserver.wait_until_serving()
        yield webserver
    finally:
        webserver.stop_serving()
        TEST_PORT += 1


@pytest.fixture()
def solara_app(solara_server):
    @contextlib.contextmanager
    def run(app: Union[solara.server.app.AppScript, str]):
        solara.server.server.solara_app.close()
        if isinstance(app, str):
            app = solara.server.app.AppScript(app)
        solara.server.server.solara_app = app
        try:
            yield
        finally:
            if app.type == solara.server.app.AppType.MODULE:
                del sys.modules[app.name]
                reload.reloader.watched_modules.remove(app.name)

            app.close()

    return run



"""
# Button
A button that can be clicked to trigger an event.
"""
from solara.kitchensink import react, sol


@react.component
def ButtonDemo():
    count, set_count = react.use_state(0)

    def increment():
        set_count(count + 1)

    with sol.VBox() as main:
        with sol.HBox():
            sol.Button(label=f"Clicked {count} times", on_click=increment, icon_name="mdi-thumb-up")
    return main


Component = sol.Button
App = ButtonDemo
app = App()



import react_ipywidgets as react
from react_ipywidgets import ipywidgets as w


@react.component
def Markdown(md_text: str):
    import markdown

    html = markdown.markdown(md_text)
    return w.HTML(value=html)


# app =
